#include "DriverSlipStream.h"
#include "Integrator.h"
#include "mylib/exception.h"
#include "mylib/function.h"
#include "mylib/SetupFileModifier.h"
#include "mylib/inifile.h"

namespace fs {
	using namespace boost::filesystem;
};

namespace {
	void findInfDirectory(std::vector<std::wstring> *result, fs::wpath const &parent_dir_path, fs::wpath const &sub_dir_path)
	{
		bool is_inf_found = false;
		fs::wdirectory_iterator it_end;
		for (fs::wdirectory_iterator it(parent_dir_path/sub_dir_path); it != it_end; ++it) {
			fs::wpath p = it->path();
			if (fs::is_directory(p)) {
				std::wstring const dir_name = p.filename();
				findInfDirectory(result, parent_dir_path, sub_dir_path/dir_name);
			} else if (!is_inf_found) {
				std::wstring const extension = fs::extension(p);
				if (wcsicmp(extension.c_str(), L".inf") == 0) {
					result->push_back(sub_dir_path.native_directory_string());
					is_inf_found = true;
				}
			}
		}
	}
};

DriverSlipStream::DriverSlipStream(std::vector<DriverSetupData> add_drivers,
                                   DriverFileDataMap const &leave_drivers,
                                   DriverFileDataMap const &del_drivers,
                                   DriverServiceData const &leave_services,
                                   DriverServiceData const &del_services,
                                   wchar_t reserved_drive,
                                   bool adapts_big_drive,
                                   bool recompresses_driver_cab,
                                   boost::filesystem::wpath const &work_dir_path,
                                   std::wstring const &i386_subdir_name,
                                   int architecture_id,
                                   uint32_t os_version)
	: add_drivers_(add_drivers)
	, leave_drivers_(leave_drivers)
	, del_drivers_(del_drivers)
	, leave_services_(leave_services)
	, del_services_(del_services)
	, reserved_drive(reserved_drive)
	, adapts_big_drive_(adapts_big_drive)
	, recompresses_driver_cab_(recompresses_driver_cab)
	, work_dir_path_(work_dir_path)
	, i386_subdir_name_(i386_subdir_name)
	, architecture_id_(architecture_id)
	, os_version_(os_version)
{
}

void DriverSlipStream::execute()
{
	{
		fs::wpath const target_path = this->work_dir_path_/this->i386_subdir_name_;
		std::auto_ptr<MyLib::SetupFileModifier> txtsetup_sif;
		std::auto_ptr<MyLib::SetupFileModifier> dosnet_inf;
		std::auto_ptr<MyLib::RegHive> setupreg_hiv;
		std::auto_ptr<MyLib::SetupFileModifier> hivesys_inf;
		std::auto_ptr<MyLib::SetupFileModifier> hivesft_inf;
		std::auto_ptr<MyLib::SetupFileModifier> hivsft32_inf;
		if (!this->del_drivers_.empty() || !this->add_drivers_.empty() || this->recompresses_driver_cab_) {
			fs::wpath const txtsetup_sif_path = target_path/L"txtsetup.sif";
			txtsetup_sif.reset(new MyLib::SetupFileModifier(txtsetup_sif_path));
			fs::wpath const dosnet_inf_path = target_path/L"dosnet.inf";
			dosnet_inf.reset(new MyLib::SetupFileModifier(dosnet_inf_path));
		}
		if (!this->del_drivers_.empty() || !this->add_drivers_.empty() || this->adapts_big_drive_) {
			fs::wpath const setupreg_hiv_path = target_path/L"setupreg.hiv";
			setupreg_hiv.reset(new MyLib::RegHive(setupreg_hiv_path));
		}
		if (!this->del_drivers_.empty() || !this->add_drivers_.empty()) {
			fs::wpath const hivesys_inf_path = target_path/L"hivesys.inf";
			hivesys_inf.reset(new MyLib::SetupFileModifier(hivesys_inf_path));
			fs::wpath const hivesft_inf_path = target_path/L"hivesft.inf";
			hivesft_inf.reset(new MyLib::SetupFileModifier(hivesft_inf_path));
			// x64
			if (this->architecture_id_ == PROCESSOR_ARCHITECTURE_AMD64 || this->architecture_id_ == PROCESSOR_ARCHITECTURE_IA64) {
				fs::wpath const hivsft32_inf_path = target_path/L"hivsft32.inf";
				hivsft32_inf.reset(new MyLib::SetupFileModifier(hivsft32_inf_path));
			}
		}
		Integrator integrator(txtsetup_sif.get(),
							  dosnet_inf.get(),
							  setupreg_hiv.get(),
							  hivesys_inf.get(),
							  this->work_dir_path_,
							  this->i386_subdir_name_,
							  this->architecture_id_,
							  this->os_version_);
		// Event Listener
		for (unsigned int i = 0; i < this->listeners.size(); ++i) {
			integrator.addEventListener(this->listeners[i]);
		}
		if (!this->del_drivers_.empty()) {
			this->fireEvent(L"ߋɓhCo𐮗Ă܂BBB");
			integrator.removeDriver(this->leave_drivers_, this->del_drivers_, this->leave_services_, this->del_services_);
			this->fireEvent(L"\n");
		}
		if (!this->add_drivers_.empty()) {
			integrator.prepareDriverIntegration();
			this->fireEvent(L"VhCoǉĂ܂BBB\n");
			std::vector<DriverSetupData>::const_iterator it1 = this->add_drivers_.begin();
			//for (size_t i = 0; i < this->add_drivers_.size(); ++i) {
			for (; it1 != this->add_drivers_.end(); ++it1) {
				DriverSetupData::DDInstallCollection const &ddic = it1->getDDInstallSectionCollection();
				DriverSetupData::DDInstallCollection::const_iterator it2 = ddic.begin();
				DriverSetupData::DescriptionCollection::const_iterator it3 = it2->second.descriptions.begin();
				this->fireEvent((it3 != it2->second.descriptions.end()) ? (L"\t" + *it3 + L"\n") : L"\tNo Description\n");
				//integrator.integrateDriver(this->add_drivers_[i]);
				integrator.integrateDriver(*it1);
			}

			this->fireEvent(L"\n");
		}
		// BigDrive
		if (this->adapts_big_drive_) {
			assert(this->os_version_ == 0x00050000);
			integrator.adaptBigDrive();
		}
		// driver.cab
		if (this->recompresses_driver_cab_) {
			this->fireEvent(L"Cabt@CĈkĂ܂BBB");
			integrator.mergeCabFiles(this->work_dir_path_.string().c_str());
			this->fireEvent(L"\n");
		}
		//integrator.flush();
		if (txtsetup_sif.get() != NULL) {
			this->fireEvent(_T("TXTSETUP.SIF CĂ܂BBB"));
			txtsetup_sif->flush();
			txtsetup_sif.reset();
			this->fireEvent(_T("\n"));
		}
		if (dosnet_inf.get() != NULL) {
			this->fireEvent(_T("DOSNET.INF CĂ܂BBB"));
			dosnet_inf->flush();
			dosnet_inf.reset();
			this->fireEvent(_T("\n"));
		}
		if (setupreg_hiv.get() != NULL) {
			this->fireEvent(L"SETUPREG.HIV CĂ܂BBB");
			setupreg_hiv.reset();
			this->fireEvent(L"\n");
		}
		if (hivesys_inf.get() != NULL) {
			this->fireEvent(_T("HIVESYS.INF CĂ܂BBB"));
			hivesys_inf->flush();
			hivesys_inf.reset();
			this->fireEvent(_T("\n"));
		}
		std::wstring device_path_add;
		{
			IniFile ini_file(*PROFILE_PATH);
			std::wstring str = ini_file.getString(L"Settings", L"driver_drive", L"");
			// hCofBNgȉċAIɑAinft@CfBNg
			std::vector<std::wstring> device_paths;
			findInfDirectory(&device_paths, this->work_dir_path_/L"$OEM$/Drivers", fs::wpath());
			std::vector<std::wstring>::const_iterator it = device_paths.begin();
			wchar_t const device_path_base[] = {L';', this->reserved_drive, L':', L'\\', L'\0'};
			for (; it != device_paths.end(); ++it) {
				device_path_add += device_path_base;
				device_path_add += *it;
			}
		}
		if (hivesft_inf.get() != NULL) {
			this->fireEvent(L"HIVESFT.INF CĂ܂BBB");
			// ̃Gg폜
			std::wstring original_device_path = L"%SystemRoot%\\inf";
			{
				MyLib::SetupFileLine::FieldCollection fields;
				fields[1] = L"HKLM";
				fields[2] = L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion";
				fields[3] = L"DevicePath";
				MyLib::SetupFileLineIterator2 it = hivesft_inf->findFirstLine(L"AddReg");
				MyLib::SetupFileLineIterator2 const it_end;
				it = hivesft_inf->findMatchLine(it, fields);
				if (it != it_end) {
					original_device_path = it->getStringField(5);
					it = hivesft_inf->eraseLine(it);
				}
				while ((it = hivesft_inf->findMatchLine(it, fields)) != it_end) {
					int n = it->getFieldCount();
					for (int i = 1; i <= n; ++i) {
						std::wstring str = it->getStringField(i);
						//int l = str.length();
					}
					it = hivesft_inf->eraseLine(it);
				}
			}
			// VGgǉ
			{
				MyLib::SetupFileLine::FieldCollection fields;
				fields[1] = L"HKLM";
				fields[2] = L"\"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\"";
				fields[3] = L"\"DevicePath\"";
				fields[4] = L"0x00020002";
				fields[5] = L'"' + original_device_path + device_path_add + L'"';
				hivesft_inf->insertLine(L"AddReg", fields);
			}
			// ̃Gg폜
			{
				MyLib::SetupFileLine::FieldCollection fields;
				fields[1] = L"HKLM";
				fields[2] = L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion";
				fields[3] = L"F6ToolDevicePath";
				MyLib::SetupFileLineIterator2 it = hivesft_inf->findFirstLine(L"AddReg");
				MyLib::SetupFileLineIterator2 it_end;
				while ((it = hivesft_inf->findMatchLine(it, fields)) != it_end) {
					it = hivesft_inf->eraseLine(it);
				}
			}
			// VGgǉ
			{
				MyLib::SetupFileLine::FieldCollection fields;
				fields[1] = L"HKLM";
				fields[2] = L"\"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\"";
				fields[3] = L"\"F6ToolDevicePath\"";
				fields[4] = L"0x00000000";
				wchar_t const f6tool_device_path[] = {L'"', this->reserved_drive, L':', L'"', L'\0'};
				fields[5] = f6tool_device_path;
				hivesft_inf->insertLine(L"AddReg", fields);
			}
			hivesft_inf->flush();
			hivesft_inf.reset();
			this->fireEvent(L"\n");
		}
		if (hivsft32_inf.get() != NULL) {
			this->fireEvent(L"HIVSFT32.INF CĂ܂BBB");
			//HKLM,"SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion","DevicePath",0x00020002,"%SystemRoot%\inf"
			// ̃Gg폜
			std::wstring original_device_path = L"%SystemRoot%\\inf";
			{
				MyLib::SetupFileLine::FieldCollection fields;
				fields[1] = L"HKLM";
				fields[2] = L"SOFTWARE\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion";
				fields[3] = L"DevicePath";
				MyLib::SetupFileLineIterator2 it = hivsft32_inf->findFirstLine(L"AddReg");
				MyLib::SetupFileLineIterator2 it_end;
				it = hivsft32_inf->findMatchLine(it, fields);
				if (it != it_end) {
					original_device_path = it->getStringField(5);
					it = hivsft32_inf->eraseLine(it);
				}
				while ((it = hivsft32_inf->findMatchLine(it, fields)) != it_end) {
					it = hivsft32_inf->eraseLine(it);
				}
			}
			// VGgǉ
			{
				MyLib::SetupFileLine::FieldCollection fields;
				fields[1] = L"HKLM";
				fields[2] = L"\"SOFTWARE\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\"";
				fields[3] = L"\"DevicePath\"";
				fields[4] = L"0x00020002";
				fields[5] = L'"' + original_device_path + device_path_add + L'"';
				hivsft32_inf->insertLine(L"AddReg", fields);
			}
			hivsft32_inf->flush();
			hivsft32_inf.reset();
			this->fireEvent(L"\n");
		}
	}
}
