#include "ISOImageCreator.h"
#include "mylib/inifile.h"
#include "mylib/function.h"
#include "mylib/ConsoleProcessExecutor.h"

extern void removeFileForcibly(boost::filesystem::wpath const &file_path);
extern void stringReplace(std::wstring *str,
                          std::wstring const &find_str,
                          std::wstring const &replace_str);

wchar_t const *ISOImageCreator::CDIMAGE_EXE_FILE_TITLE = L"cdimage.exe";
wchar_t const *ISOImageCreator::MKISOFS_EXE_FILE_TITLE = L"mkisofs.exe";
wchar_t const *ISOImageCreator::BOOTSECT_BIN_FILE_TITLE = L"bootsect.bin";

namespace
{
	// convertAllFilenamesToUpper
	//  w肳ꂽfBNgȉ̃t@CEfBNgׂđ啶ɂ
	void convertAllFilenamesToUpper(boost::filesystem::wpath const &dir_path)
	{
		namespace fs = boost::filesystem;
		fs::wdirectory_iterator it_end;
		for (fs::wdirectory_iterator it(dir_path); it != it_end; ++it) {
			fs::wpath const file_path = it->path();
			std::wstring const file_name_str = file_path.filename();
			size_t file_name_len = file_name_str.length();
			boost::scoped_array<wchar_t> buf(new wchar_t[file_name_len + 1]);
			::lstrcpy(buf.get(), file_name_str.c_str());
			::CharUpper(buf.get());
			fs::wpath const upper_path = dir_path/buf.get();
			fs::rename(file_path, upper_path);
			//::MoveFile(file_path.string().c_str(), upper_path.string().c_str());
			if (fs::is_directory(it->status())) {
				// ċNĂяo
				convertAllFilenamesToUpper(upper_path);
			}
		}
	}
};

void stringReplace(std::wstring *str, std::wstring const &find_str, std::wstring const &replace_str)
{
	std::wstring::size_type pos;
	for (pos = str->find(find_str); pos != std::wstring::npos; pos = str->find(find_str, replace_str.length() + pos)) {
		str->replace(pos, find_str.length(), replace_str);
	}
}

namespace {
static size_t const BOOT_IMAGE_SIZE = 2048;

bool isBootImage(unsigned char const *ptr0)
{
	// think of boot image if contains "CDBOOT:"
	static char const *CDBOOT = "CDBOOT:";
	static size_t const CDBOOT_LEN = strlen(CDBOOT);
	
	size_t search_range = BOOT_IMAGE_SIZE;
	unsigned char const *ptr =
		reinterpret_cast<unsigned char const*>(std::memchr(ptr0, CDBOOT[0], search_range));
	while (ptr != NULL) {
		if (std::memcmp(ptr, CDBOOT, CDBOOT_LEN) == 0) {
			return true;
		}
		search_range = BOOT_IMAGE_SIZE - (ptr - ptr0 + 1);
		ptr = reinterpret_cast<unsigned char const*>(std::memchr(ptr + 1, CDBOOT[0], search_range));
	}
	return false;
}

unsigned char const *getBootImageStartPoint(unsigned char *buf, size_t buf_size)
{
	if (buf_size > BOOT_IMAGE_SIZE) {
		static unsigned char const SIGNATURE[] = {0x55, 0xAA};
		static size_t const SIGNATURE_SIZE = _countof(SIGNATURE);

		unsigned char const *ptr = NULL;
		size_t const search_range0 = buf_size - (BOOT_IMAGE_SIZE - SIGNATURE_SIZE);
		size_t search_range = search_range0;
		ptr = reinterpret_cast<unsigned char const*>(std::memchr(buf + (BOOT_IMAGE_SIZE - SIGNATURE_SIZE), SIGNATURE[0], search_range));
		while (ptr != NULL) {
			size_t n1 = ptr - buf + 1;
			int nn1 = n1 % 16;
			if (ptr[1] == SIGNATURE[1] && nn1 == 15) {
				unsigned char const *ptr1 = ptr - (BOOT_IMAGE_SIZE - SIGNATURE_SIZE);
				if (isBootImage(ptr1)) {
					return ptr1;
				}
			}
			search_range = search_range0 + n1;
			ptr = reinterpret_cast<unsigned char const*>(std::memchr(ptr + 1, SIGNATURE[0], search_range));
		}
	}
	return NULL;
}
};  // namespace

/////////////////////////////////////////////////////////////////////////////

ISOImageCreator::ISOImageCreator(boost::filesystem::wpath const &iso_image_path,
                                 boost::filesystem::wpath const &target_dir_path,
                                 boost::filesystem::wpath const &source_path,
                                 std::wstring const &i386_subdir_name,
                                 std::wstring const &volume_label)
: iso_image_path_(iso_image_path)
, target_dir_path_(target_dir_path)
{
	namespace fs = boost::filesystem;

	std::wstring cmdline;
	std::wstring mkisofs_file_title;

	// u[gC[W̏
	fs::wpath const bootsect_bin_path = *BIN_DIR_PATH/BOOTSECT_BIN_FILE_TITLE;
	if (!fs::exists(bootsect_bin_path)) {
		// SPCMDCON.SYS u[gC[W𔲂o
		fs::wpath spcmdcon_sys_path = source_path/i386_subdir_name/L"SPCMDCON.SYS";
		fs::ifstream ifs(spcmdcon_sys_path, std::ios::in | std::ios::binary);
		if (!ifs) {
			//std::wprintf(L"file open error\n");
			//throw
		}
		size_t const s = ifs.seekg(0, std::ios::end).tellg();
		ifs.seekg(0, std::ios::beg);
		boost::scoped_array<unsigned char> buf(new unsigned char[s]);
		ifs.read(reinterpret_cast<char*>(buf.get()), s);

		unsigned char const *ptr = getBootImageStartPoint(buf.get(), s);
		if (ptr != NULL) {
			fs::ofstream ofs(bootsect_bin_path, std::ios::out | std::ios::binary);
			if (!ofs) {
				//throw
			}
			ofs.write(reinterpret_cast<char const*>(ptr), BOOT_IMAGE_SIZE);
		}
	}

	IniFile ini_file(*PROFILE_PATH);
	// CDIMAGED
	fs::wpath const cdimage_exe_path = *BIN_DIR_PATH/CDIMAGE_EXE_FILE_TITLE;
	if (fs::exists(cdimage_exe_path)) {
		this->mkisofs_file_title_ = CDIMAGE_EXE_FILE_TITLE;
		// CDIMAGE.EXE ̃IvV
		std::wstring options = ini_file.getString(L"Settings", L"cdimage_options", L"");
		if (options.empty()) {
			options = L"-l\"##VOLUME_LABEL##\" -h -n -m -o -x -b\"##BOOT_IMAGE##\" \"##FOLDER##\" \"##ISO_IMAGE##\"";
			ini_file.setString(L"Settings", L"cdimage_options", options);
		}
		stringReplace(&options, L"##VOLUME_LABEL##", volume_label);
		stringReplace(&options, L"##BOOT_IMAGE##", bootsect_bin_path.string());
		stringReplace(&options, L"##FOLDER##", target_dir_path.string());
		stringReplace(&options, L"##ISO_IMAGE##", iso_image_path.string());
		this->cmdline_ = L"\"" + cdimage_exe_path.string() + L"\" " + options;
	} else {
		fs::wpath const mkisofs_exe_path = *BIN_DIR_PATH/MKISOFS_EXE_FILE_TITLE;
		//if (!fs::exists(mkisofs_exe_path)) {
			// mkisofs.exe is not found
			//throw
		//}
		this->mkisofs_file_title_ = MKISOFS_EXE_FILE_TITLE;
		// mkisofs.exe ̃IvV
		std::wstring options = ini_file.getString(L"Settings", L"mkisofs_options", L"");
		if (options.empty()) {
			options = L"-iso-level 4 -duplicates-once -b \"##BOOT_IMAGE##\" -hide boot.catalog -hide \"##BOOT_IMAGE##\" -no-emul-boot -V \"##VOLUME_LABEL##\" -A \"CDIMAGE 2.50 (10/12/2000 TM)\" -publisher \"MICROSOFT CORPORATION\" -p \"MICROSOFT CORPORATION, ONE MICROSOFT WAY, REDMOND WA 98052, (425) 882-8080\" -volset \"##VOLUME_LABEL##\" -o \"##ISO_IMAGE##\" \"##FOLDER##\"";
			ini_file.setString(L"Settings", L"mkisofs_options", options);
		}
		stringReplace(&options, L"##VOLUME_LABEL##", volume_label);
		stringReplace(&options, L"##BOOT_IMAGE##", BOOTSECT_BIN_FILE_TITLE);
		stringReplace(&options, L"##FOLDER##", target_dir_path.string());
		stringReplace(&options, L"##ISO_IMAGE##", iso_image_path.string());
		this->cmdline_ = L"\"" + mkisofs_exe_path.string() + L"\" " + options;
	}
}

void ISOImageCreator::execute()
{
	// t@C̑啶
	this->fireEvent(L"ׂẴt@C啶ɂĂ܂BBB");
	convertAllFilenamesToUpper(this->target_dir_path_);
	this->fireEvent(L"\n");

	// bootsect.biñRs[
	if (this->mkisofs_file_title_ == MKISOFS_EXE_FILE_TITLE) {
		namespace fs = boost::filesystem;
		fs::wpath dst_path = this->target_dir_path_/BOOTSECT_BIN_FILE_TITLE;
		removeFileForcibly(dst_path);
		fs::wpath src_path = *BIN_DIR_PATH/BOOTSECT_BIN_FILE_TITLE;
		if (!::CopyFile(src_path.string().c_str(), dst_path.string().c_str(), FALSE)) {
			//throw
		}
	}

	// ISOC[WfBNg邩mF
	namespace fs = boost::filesystem;
	fs::wpath iso_image_dir_path = this->iso_image_path_;
	iso_image_dir_path.remove_filename();
	if (!fs::exists(iso_image_dir_path)) {
		fs::create_directories(iso_image_dir_path);
	}

	this->fireEvent(this->cmdline_ + L'\n');
	MyLib::ConsoleProcessExecutor mkisofs_exe;
	mkisofs_exe.run(this->cmdline_);
	unsigned int status;
	while ((status = mkisofs_exe.getStatus(100)) != WAIT_ABANDONED) {
		std::vector<char> buf = mkisofs_exe.readStdOut();
		if (!buf.empty()) {
			buf.push_back('\0');
			this->fireEvent(MyLib::convertFromSJIS(&buf[0]));
		}
		buf = mkisofs_exe.readStdErr();
		if (!buf.empty()) {
			buf.push_back('\0');
			this->fireEvent(MyLib::convertFromSJIS(&buf[0]));
		}
		if (status == WAIT_OBJECT_0) break;
	}
	const unsigned int exit_code = mkisofs_exe.getExitCode();
	if (exit_code != 0) {
		TCHAR buf[64];
		::wsprintf(buf, _T("%s returned %d."), this->mkisofs_file_title_.c_str(), exit_code);
		throw std::runtime_error(MyLib::convertUTF16ToUTF8(buf));
	}
}
