﻿#include "SourceSetting.h"

#include "mylib/function.h"
#include "mylib/InfFile.h"

namespace fs = boost::filesystem;

namespace {
// findSetupSubDirectory
//  TXTSETUP.SIFなどが存在するディレクトリを探す
std::wstring findSetupSubDirectory(fs::wpath const &root)
{
	fs::wdirectory_iterator it_end;
	for (fs::wdirectory_iterator it(root); it != it_end; ++it) {
		fs::wpath p = it->path();
		if (fs::is_directory(p)) {
			fs::wpath const txtsetup_sif_path = p/L"txtsetup.sif";
			fs::wpath const dosnet_inf_path = p/L"dosnet.inf";
			//fs::wpath const svcpack_inf_path = 
			fs::wpath const setupreg_hiv_path = p/L"setupreg.hiv";
			if (fs::exists(txtsetup_sif_path)
			 && fs::exists(dosnet_inf_path)
			 && fs::exists(setupreg_hiv_path)) {
				return p.filename();
			}
		}
	}
	return L"";
}

// CPUアーキテクチャ(x86,AMD64,IA64)を取得
int getArchitectureIDFromTxtsetupSif(MyLib::InfFile &txtsetup_sif)
{
	using namespace MyLib;

	InfContextIterator const it_end;
	InfContextIterator it =
		txtsetup_sif.findFirstLine(L"SetupData", L"Architecture");
	if (it != it_end) {
		std::wstring const str = it->getStringField(1);
		if (compareStringNoCase(str, L"i386") == 0) {
			return PROCESSOR_ARCHITECTURE_INTEL;
		} else if (compareStringNoCase(str, L"AMD64") == 0) {
			return PROCESSOR_ARCHITECTURE_AMD64;
		} else if (compareStringNoCase(str, L"IA64") == 0) {
			return PROCESSOR_ARCHITECTURE_IA64;
		}
	}
	return -1;
}

// OSバージョンを取得
uint32_t getOSVersionFromTxtsetupSif(MyLib::InfFile &txtsetup_sif)
{
	using namespace MyLib;

	InfContextIterator const it_end;
	uint32_t major_ver, minor_ver;
	InfContextIterator it1 =
		txtsetup_sif.findFirstLine(L"SetupData", L"MajorVersion");
	if (it1 != it_end) {
		major_ver = it1->getIntField(1);
	} else {
		//throw std::runtime_error("TXTSETUP.SIF is invalid");
		return 0;
	}
	MyLib::InfContextIterator it2 =
		txtsetup_sif.findFirstLine(L"SetupData", L"MinorVersion");
	if (it2 != it_end) {
		minor_ver = it2->getIntField(1);
	} else {
		//throw std::runtime_error("TXTSETUP.SIF is invalid");
		return 0;
	}
	return (major_ver << 16) | minor_ver;
}

// Product type
int getProductTypeFromTxtsetupSif(MyLib::InfFile &txtsetup_sif)
{
	using namespace MyLib;

	InfContextIterator const it_end;
	InfContextIterator it =
		txtsetup_sif.findFirstLine(L"SetupData", L"ProductType");
	if (it != it_end) {
		return it->getIntField(1);
	}
	return -1;
}

int estimateCurrentSPNumber(fs::wpath const &source_path)
{
	fs::wpath const wild_card = source_path/L"SP?.CAT";
	CFindFile finder;
	int result = 0;
	if (finder.FindFile(wild_card.string().c_str())) {
		do {
			TCHAR buf[MAX_PATH];
			finder.GetFileName(buf, MAX_PATH);
			if (::isdigit(buf[2])) {
				int n = buf[2] - _T('0');
				if (n > result) {
					result = n;
				}
			}
		} while (finder.FindNextFile());
	}
	return result;
}

bool isDiskWritable(fs::wpath const &source_path)
{
	fs::wpath const complete_p = fs::system_complete(source_path);
	std::wstring const complete_p_str = complete_p.string();
	wchar_t const root_p[] = {complete_p_str[0], ':', '\\', '\0'};
	if (::GetDriveType(root_p) == DRIVE_CDROM) {
		return false;
	}

	wchar_t const str[] = {'\\', '\\', '.', '\\', complete_p_str[0], ':', '\0'};
	MyLib::AutoHandleI h = ::CreateFile(str, 0,
	                                    FILE_SHARE_READ | FILE_SHARE_WRITE,
	                                    NULL, OPEN_EXISTING,
	                                    FILE_ATTRIBUTE_NORMAL, NULL);
	if (h.isInvalid()) {
		throw MY_WIN32_EXCEPTION_LAST_ERROR;
	}
	DWORD bytes_returned;
	BOOL ret = ::DeviceIoControl(h.get(),
	                             IOCTL_DISK_IS_WRITABLE,
	                             NULL, 0, NULL, 0,
	                             &bytes_returned, NULL);
	DWORD err_code = ::GetLastError();
	if (!ret && err_code != ERROR_WRITE_PROTECT) {
		throw MY_WIN32_EXCEPTION(err_code);
	}
	return ret != FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// boot image
/////////////////////////////////////////////////////////////////////////////

static size_t const BOOT_IMAGE_SIZE = 2048;

bool isBootImage(unsigned char const *ptr0)
{
	// think of boot image if contains "CDBOOT:"
	static char const *CDBOOT = "CDBOOT:";
	static size_t const CDBOOT_LEN = strlen(CDBOOT);
	
	size_t search_range = BOOT_IMAGE_SIZE;
	unsigned char const *ptr =
		reinterpret_cast<unsigned char const*>(std::memchr(ptr0, CDBOOT[0], search_range));
	while (ptr != NULL) {
		if (std::memcmp(ptr, CDBOOT, CDBOOT_LEN) == 0) {
			return true;
		}
		search_range = BOOT_IMAGE_SIZE - (ptr - ptr0 + 1);
		ptr = reinterpret_cast<unsigned char const*>(std::memchr(ptr + 1, CDBOOT[0], search_range));
	}
	return false;
}

unsigned char const *getBootImageStartPoint(unsigned char *buf, size_t buf_size)
{
	if (buf_size > BOOT_IMAGE_SIZE) {
		static unsigned char const SIGNATURE[] = {0x55, 0xAA};
		static size_t const SIGNATURE_SIZE = _countof(SIGNATURE);

		unsigned char const *ptr = NULL;
		size_t const search_range0 = buf_size - (BOOT_IMAGE_SIZE - SIGNATURE_SIZE);
		size_t search_range = search_range0;
		ptr = reinterpret_cast<unsigned char const*>(std::memchr(buf + (BOOT_IMAGE_SIZE - SIGNATURE_SIZE), SIGNATURE[0], search_range));
		while (ptr != NULL) {
			size_t n1 = ptr - buf + 1;
			int nn1 = n1 % 16;
			if (ptr[1] == SIGNATURE[1] && nn1 == 15) {
				unsigned char const *ptr1 = ptr - (BOOT_IMAGE_SIZE - SIGNATURE_SIZE);
				if (isBootImage(ptr1)) {
					return ptr1;
				}
			}
			search_range = search_range0 + n1;
			ptr = reinterpret_cast<unsigned char const*>(std::memchr(ptr + 1, SIGNATURE[0], search_range));
		}
	}
	return NULL;
}

void saveBootImage(fs::wpath const &spcmdcon_sys_path, fs::wpath const &bootsect_bin_path)
{
	// SPCMDCON.SYS からブートイメージを抜き出す
	fs::ifstream ifs(spcmdcon_sys_path, std::ios::in | std::ios::binary);
	if (!ifs) {
		throw MY_STD_C_FILE_EXCEPTION(spcmdcon_sys_path.string(), EINVAL);
	}
	// ファイル全体をバッファに読み込む
	size_t const s = ifs.seekg(0, std::ios::end).tellg();
	ifs.seekg(0, std::ios::beg);
	boost::scoped_array<unsigned char> buf(new unsigned char[s]);
	ifs.read(reinterpret_cast<char*>(buf.get()), s);
	// ブートイメージの部分を探し、ファイルに保存
	unsigned char const *ptr = getBootImageStartPoint(buf.get(), s);
	if (ptr != NULL) {
		fs::ofstream ofs(bootsect_bin_path, std::ios::out | std::ios::binary);
		if (!ofs) {
			throw MY_STD_C_FILE_EXCEPTION(bootsect_bin_path.string(), EINVAL);
		}
		ofs.write(reinterpret_cast<char const*>(ptr), BOOT_IMAGE_SIZE);
	} else {
		throw MY_STD_C_FILE_EXCEPTION(spcmdcon_sys_path.string(), EINVAL);
	}
}
};

/////////////////////////////////////////////////////////////////////////////

//SourceSetting::SourceSetting(std::wstring const &setup_subdir_name,
//                             int architecture_id,
//                             uint32_t os_version,
//                             int sp_number)
//	: setup_subdir_name(setup_subdir_name)
//	, architecture_id(architecture_id)
//	, os_version(os_version)
//	, sp_number(sp_number)
//{
//}

std::wstring SourceSetting::getI386SubdirName() const
{
	return this->i386_subdir_name;
}

int SourceSetting::getArchitectureID() const
{
	return this->architecture_id;
}

uint32_t SourceSetting::getOSVersion() const
{
	return this->os_version;
}

int SourceSetting::getProductType() const
{
	return this->product_type;
}

int SourceSetting::getCurrentSPNumber() const
{
	return this->sp_number;
}

bool SourceSetting::isReadOnly() const
{
	return this->is_read_only;
}

/////////////////////////////////////////////////////////////////////////////

DirectorySourceSetting::DirectorySourceSetting(fs::wpath const &source_path)
{
	// TXTSETUP.SIFなどを探す
	std::wstring const setup_subdir_name = findSetupSubDirectory(source_path);
	if (setup_subdir_name.empty()) {
		throw std::runtime_error("TXTSETUP.SIF or other file is not found.");
	}

	boost::filesystem::wpath const txtsetup_sif_path =
		source_path/setup_subdir_name/L"txtsetup.sif";
	MyLib::InfFile txtsetup_sif(txtsetup_sif_path.string().c_str());
	// CPUアーキテクチャ(x86,AMD64,IA64)を取得
	this->architecture_id = getArchitectureIDFromTxtsetupSif(txtsetup_sif);
	// OSバージョンを取得
	this->os_version = getOSVersionFromTxtsetupSif(txtsetup_sif);
	// Product type
	this->product_type = getProductTypeFromTxtsetupSif(txtsetup_sif);
	// SP取得
	this->sp_number = estimateCurrentSPNumber(source_path/setup_subdir_name);

	// 反映させる
	this->source_path = source_path;
	this->i386_subdir_name = setup_subdir_name;

	this->is_read_only = !isDiskWritable(source_path);
}

void DirectorySourceSetting::copySource(fs::wpath const &/*dest_path*/)
{

}

void DirectorySourceSetting::createDiscBootSector(fs::wpath const &boot_sector_path)
{
	fs::wpath const spcmdcon_sys_path = source_path/i386_subdir_name/L"SPCMDCON.SYS";
	saveBootImage(spcmdcon_sys_path, boot_sector_path);
}

boost::filesystem::wpath DirectorySourceSetting::getSourcePath() const
{
	return this->source_path;
}

/////////////////////////////////////////////////////////////////////////////

DiscSourceSetting::DiscSourceSetting(fs::wpath const &source_path)
: DirectorySourceSetting(source_path)
{
	fs::wpath const complete_p = fs::system_complete(source_path);
	std::wstring const complete_p_str = complete_p.string();
	wchar_t const root_p[] = {complete_p_str[0], ':', '\\', '\0'};
	wchar_t volume_label[MAX_PATH + 1];
	// ボリュームラベルを取得する
	// SetErrorMode()は「ディスクを入れろ」メッセージボックスを抑止するため
	// 実際は出てこないけど、MSDNに書いてあるので念のため
	unsigned int const prev_error_mode = ::SetErrorMode(SEM_FAILCRITICALERRORS);
	bool const ret = (::GetVolumeInformation(
		root_p, volume_label, MAX_PATH + 1, NULL, NULL, NULL, NULL, 0) != 0);
	DWORD const err_code = (ret ? 0 : ::GetLastError());
	::SetErrorMode(prev_error_mode);
	if (!ret) {
		throw MY_WIN32_EXCEPTION(err_code);
	}
	this->volume_label = volume_label;

	//// TXTSETUP.SIFなどを探す
	//std::wstring const setup_subdir_name = findSetupSubDirectory(source_path);
	//if (setup_subdir_name.empty()) {
	//	throw std::runtime_error("TXTSETUP.SIF or other file is not found.");
	//}

	//boost::filesystem::wpath const txtsetup_sif_path =
	//	source_path/setup_subdir_name/L"txtsetup.sif";
	//MyLib::InfFile txtsetup_sif(txtsetup_sif_path.string().c_str());
	//// CPUアーキテクチャ(x86,AMD64,IA64)を取得
	//this->architecture_id = getArchitectureIDFromTxtsetupSif(txtsetup_sif);
	//// OSバージョンを取得
	//this->os_version = getOSVersionFromTxtsetupSif(txtsetup_sif);
	//// SP取得
	//this->sp_number = estimateCurrentSPNumber(source_path/i386_subdir_name);

	//// 反映させる
	//this->source_path = source_path;
	//this->i386_subdir_name = setup_subdir_name;

}

std::wstring DiscSourceSetting::getVolumeLabel() const
{
	return this->volume_label;
}
